#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <pthread.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <assert.h>

#include "queue.h"
#include "count.h"
#include "containers.h"
#include "config.h"
#include "safe_write.h"

#define _SMMAPD_C_
#include "smmapd.h"


#ifdef SUNOS
#include "sunos_comp.h"
#endif


#define CFG_SECTION_GLOBAL "global"
#define CFG_NAME_ADDRESS "address"
#define CFG_NAME_PORT "port"


#define DEFAULT_PORT "8888"

#define CVS_ID "$Id$"
#define RELEASE_ID "0.9"


struct networkerThread_s {
  int       result;
  struct sockaddr_in clientAddr; 
  socklen_t clientAddrLen;
  int       clientSock;
  pthread_t pthread;
};

typedef struct networkerThread_s networkerThread_t;


ht_queue_t terminated_networker_queue;
pthread_t cleanerThread;
count_t thread_counter;

cfg_t *cfg;


void * cleaner(void * arg) {
  networkerThread_t *t;
  int joinRes;

  pthread_detach(pthread_self());

  while (1) {
    syslog(LOG_DEBUG, "cleaner: Waiting for terminated networker, running threads: %d", 
	   count_get(&thread_counter));
    t = (networkerThread_t*) queue_get_wait(&terminated_networker_queue);
    count_dec(&thread_counter);

    syslog(LOG_DEBUG, "cleaner: Networker serving %d (address: %s, port: %d), result %d",
	   t->clientSock, inet_ntoa(t->clientAddr.sin_addr),
	   ntohs(t->clientAddr.sin_port), t->result);

    joinRes = pthread_join(t->pthread, NULL);
    syslog(LOG_DEBUG, "cleaner: joinRes %d", joinRes);

    free(t);
    t = NULL;
  }
}



	       
#define OUTPUT_BUFSIZE ANSWER_BUFSIZE+25 /* a bit more than answer for the result text */
void * networker(void * arg) {

  char buffer[BUFSIZE+1];
  char *input;
  char answer[ANSWER_BUFSIZE+1];
  char output[OUTPUT_BUFSIZE+1];
  const char *answer_ptr;
  const char *result_text;
  int err, cnt, i = 0;
  unsigned int len;
  int dispatcher_result;
  container_handle_t *container_handle;


  networkerThread_t * thread = (networkerThread_t *) arg;
  thread->pthread = pthread_self();

  syslog(LOG_DEBUG, "networker: serving %d (address: %s, port: %d) started",
	 thread->clientSock, inet_ntoa(thread->clientAddr.sin_addr),
	 ntohs(thread->clientAddr.sin_port));

  containers_setup(&container_handle);
  syslog(LOG_DEBUG, "networker: got a container handle");


  while(0 != (cnt = read(thread->clientSock, buffer, BUFSIZE))) {
    /* temporary, until netstring is introduced */
    if ((buffer[cnt-1] == 10) && (buffer[cnt-2] == 13)) /* ^M */
      buffer[cnt-2] = '\0';
    else
      buffer[cnt] = '\0';

    len = strtol(buffer, &input, 10);
    if ((0 == len) && (buffer == input)) {
      syslog(LOG_DEBUG, "networker: netstring unparsable, no length found: %s", buffer);
      dispatcher_result = SMM_NETSTRING_UNPARSABLE;
    } else if (*input++ != ':') {
      syslog(LOG_DEBUG, "networker: netstring unparsable, no colon found: %s", buffer);
      dispatcher_result = SMM_NETSTRING_UNPARSABLE;
    } else if (strlen(input)-1 != len) {
      syslog(LOG_DEBUG, "networker: netstring unparsable, length does not match: %s", buffer);
      dispatcher_result = SMM_NETSTRING_UNPARSABLE;
    } else if (input[strlen(input)-1] != ',') {
      syslog(LOG_DEBUG, "networker: netstring unparsable, no terminating comma: %s", buffer);
      dispatcher_result = SMM_NETSTRING_UNPARSABLE;
    } else {
      input[strlen(input)-1] = '\0'; /* strip off the comma */
      
      answer[0] = '\0';
      dispatcher_result = containers_dispatcher(container_handle, input, answer);
      syslog(LOG_DEBUG, "networker: dispatcher result: %d, answer: %s", dispatcher_result, answer);
    }

    result_text = (dispatcher_result > SMM_MAX_NUM) ?  
      T_SMM_RESULTS[0] : 
      T_SMM_RESULTS[dispatcher_result];

    answer_ptr  = (dispatcher_result > SMM_MAX_NUM) ?  
      T_SMM_RESULT_INFOS[0] : 
      T_SMM_RESULT_INFOS[dispatcher_result];

    if (NULL == answer_ptr)
      answer_ptr = answer;

    snprintf(output, OUTPUT_BUFSIZE, "%d:%s%s%s,", 
	     strlen(result_text) + strlen(answer_ptr) + ((strlen(answer_ptr) > 0) ? 1 : 0),
	     result_text,
	     (strlen(answer_ptr) > 0) ? " " : "",
	     answer_ptr);

    safe_write(thread->clientSock, output, strlen(output));
  }

  close(thread->clientSock);

  containers_destroy(container_handle);
  syslog(LOG_DEBUG, "networker: destroyed the container handle");


  thread->result = 1;
  queue_put(&terminated_networker_queue, thread);
}
    

int server() {
  int serverSock;
  int res;
  struct sockaddr_in servAddr;
  struct sockaddr_in clientAddr;
  socklen_t clientAddrLen = sizeof(clientAddr);
  int clientSock;
  networkerThread_t *thread;
  pthread_t tid;
  int optval = 1;

  char *cfg_address, *cfg_port;

  serverSock = socket(AF_INET, SOCK_STREAM, 0);
  if (-1 == serverSock) {
    syslog(LOG_ERR, "server: failure when creating server socket");
    return -1;
  }

  res = setsockopt(serverSock, SOL_SOCKET, SO_REUSEADDR, (void*)&optval, sizeof(optval));
  if (-1 == res) {
    syslog(LOG_ERR, "server: failure when setting socket options");
    return -1;
  }


  memset(&servAddr, 0, sizeof(servAddr));
  servAddr.sin_family = AF_INET;
  
  cfg_address = findcfg(cfg, CFG_SECTION_GLOBAL, CFG_NAME_ADDRESS);
  servAddr.sin_addr.s_addr = (NULL == cfg_address) ? htonl(INADDR_ANY) : inet_addr(cfg_address);

  cfg_port = findcfgx(cfg, CFG_SECTION_GLOBAL, CFG_NAME_PORT, DEFAULT_PORT);
  servAddr.sin_port = htons(atoi(cfg_port));

  res = bind(serverSock, (struct sockaddr *) &servAddr, sizeof(servAddr));
  if (-1 == res) {
    syslog(LOG_ERR, "server: failure when binding on server socket, errno: %d, errtxt: %s",
	   errno, strerror(errno));
    return -1;
  }

  res = listen(serverSock, 32);
  if (-1 == res) {
    syslog(LOG_ERR, "server: failure when listening on server socket, errno: %d, errtxt: %s",
	   errno, strerror(errno));
    return -1;
  }


  count_init(&thread_counter);
  queue_init(&terminated_networker_queue);
  pthread_create(&cleanerThread, NULL, &cleaner, NULL); 


  while (1) {
    syslog(LOG_DEBUG, "server: Waiting for connection");
    
    clientSock = accept(serverSock, (struct sockaddr *) &clientAddr, &clientAddrLen);
    syslog(LOG_DEBUG, "server: Got a connection %d", clientSock);

    if (-1 == clientSock) {
      syslog(LOG_ERR, "server: failure when accepting connection");
      return -1;
    }

    
    thread = (networkerThread_t *) malloc(sizeof(networkerThread_t)+10);
    if (NULL == thread) {
      syslog(LOG_ERR, "server: unable to alloc memory for networker");
      break;
    }
	     

    thread->clientSock = clientSock;
    thread->result = 0;
    thread->clientAddr = clientAddr;
    thread->clientAddrLen = clientAddrLen;

    count_inc(&thread_counter);
    res = pthread_create(&tid, NULL, &networker, thread);
    if (0 != res) {
      syslog(LOG_ERR, "server: unable to start networker thread");
      free(thread);
      count_dec(&thread_counter);
      break;
    }
  }

  return 0;
}

void usage() {
  printf("smmapd [-F] [-p pidfile] [-f cfgfile] [-h]\n");
  printf("  -F   run in foreground, overwrite smmapd.ini setting\n");
  printf("  -p   pidfile, overwrite smmapd.ini setting\n");
  printf("  -f   cfgfile\n");
  printf("  -v   version and release info\n");
  printf("  -h   print this help\n\n");
}

int main(int argc, char **argv) {
  #define DEFAULT_SMMAPD_INI "smmapd.ini"

  int err, pid, do_fork, foreground=0, c;
  char *pid_file=NULL, *cfg_file=NULL;
  
  FILE *f;

  while ((c = getopt(argc, argv,"Fp:f:hv")) != -1)
     switch (c) {
     case 'F': 
       foreground = 1;
       break;
     case 'f': 
       cfg_file = strdup(optarg);
       break;
     case 'p': 
       pid_file = strdup(optarg);
       break;
     case 'v': 
       printf("\nsmmapd - Wolfgang Hottgenroth <woho@hottis.de>\n");
       printf("  cvs_id: %s\n", CVS_ID);
       printf("  release_id: %s\n\n", RELEASE_ID);
       exit(1);
     case 'h':
     default:
       usage();
       exit(1);
     }



  openlog("smmapd", LOG_PID, LOG_LOCAL2);
  syslog(LOG_INFO, "main: started");
  cfg = readcfg((cfg_file == NULL) ? DEFAULT_SMMAPD_INI : cfg_file);
  if (0 != (err = register_all())) {
    syslog(LOG_ERR, "main: register_all fails: %d", err);
    exit(1);
  }

  do_fork = atoi(findcfgx(cfg, "global", "do_fork", "1"));
  pid = ((1 == do_fork) && (0 == foreground)) ? fork() : 0;

  if (-1 == pid) {
    syslog(LOG_ERR, "Error when forking smmapd: %d, %s", errno, strerror(errno));
    exit(1);
  } else if (0 != pid) {
    syslog(LOG_INFO, "smmapd successfully forked, child's pid is %d", pid);
    if (NULL == pid_file)
      pid_file = findcfgx(cfg, "global", "pid_file", "smmapd.pid");
    if ((f = fopen(pid_file, "w")) == NULL) {
      syslog(LOG_ERR, "unable to open pid_file %s for pid %d", pid_file, pid);
    } else {
      fprintf(f, "%d", pid);
      fclose(f);
    }
    exit(0);
  }


  if (0 != (err = server())) {
    syslog(LOG_ERR, "main: server fails: %d", err);
    exit(1);
  }
  syslog(LOG_INFO, "main: finished");
  closelog();
}
